# bash-crypto: crypto related bash library
# Copyright (C)  2019 Heinz Deinhart
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# bash-crypto version 0.2

#@@ Error Messages

#@ error-codes (bash-crypto)
# <50 .. reserved for base-base error-codes
#  9x .. errors local to each function

#@@ Message-Digest Functions

## helper to create alphabet reverse lookup table
# function alphabet_rev() {
#    local i t
#    local -A ra
#
#    for (( i=0; i < ${#a[*]}; i++ ))
#    do
#        printf -v t "%i" "'${a[$i]}"
#        eval "ra[$t]=\"$i\""
#    done
#
#    declare -p ra
# }

#@ bcr_digest_read <ret-vn> <digest>
# runs a digest against input data read from stdin
# where digest can be: md5|sha(1|224|256|384|512)|ripemd160|whirlpool
# errors: 2, 3, 4, 91 (openssl not found), 92 (openssl error)
function bcr_digest_read() {
    (( $# == 2 )) || {
	bb_ierr "bcr_digest_read called with $# instead of 2 args"
	return 2
    }

    local _ret_vn="$1"
    local _digest="$2"  

    bb_validate_vn "bcr_digest_read:ret-vn" "$_ret_vn" || return 4

    [[ "$_digest" =~ \
	^(md5|sha(1|224|256|384|512)|ripemd160|whirlpool)$ ]] || {
	bb_msg err "bcr_digest_read: illegal digest specified: $_digest"
	return 4
    }

    bb_executables_in_path bcr_digest_read openssl || return 91

    local _hash
    _hash=$(openssl dgst "-$_digest") || return 92
    printf -v "$_ret_vn" "%s" "${_hash##* }"
}

#@ bcr_digest_of_hex <ret-vn> <digest> <hex-message>
# runs a digest against a hex encoded message
# where digest can be: md5|sha(1|224|256|384|512)|ripemd160|whirlpool
# errors: 2, 3, 4, 91 (openssl not found), 92 (openssl error)
function bcr_digest_of_hex() {
    (( $# == 3 )) || {
	bb_ierr "bcr_digest_of_hex called with $# instead of 3 args"
	return 2
    }

    local _ret_vn="$1"
    local _digest="$2"
    local _hex_data="$3"

    bb_validate_vn "bcr_digest_of_hex:ret-vn" "$_ret_vn" || return 4

    bcr_digest_read "$_ret_vn" "$_digest" < <(bb_hex_write "$_hex_data") || \
	return "$?"
}

#@@ Base32 En/Decoding (RFC 4548)

#@ bcr_base32_encode_of_hex <ret-vn> <hex-data> [default|exthex|bech32]
# encodes hex representation of binary data to base32 string
# https://tools.ietf.org/html/rfc4648
# note: bech32 is untested
# errors: 2, 3, 91 (illegal char in hex), 92 (hex is of odd length)
function bcr_base32_encode_of_hex() {
    (( $# == 2 || $# == 3 )) || {
	bb_ierr "bcr_base32_encode_of_hex called with $#" \
	    "instead of 2 or 3 args"
	return 2
    }

    local _ret_vn="$1"
    local _in_hex="$2"
    local _mode="${3-default}"

    bb_validate_vn "bcr_base32_encode_of_hex:ret-vn" "$_ret_vn" || return 3
    bb_hex_validate "bcr_base32_encode_of_hex:hex-data" "$_in_hex" || \
	return 91

    # choose selected alphabet
    local -a _alphabet
    case "$_mode" in
	default)
	    _alphabet=(A B C D E F G H I J K L M N O P Q R S T U V W X \
		Y Z 2 3 4 5 6 7);;
	exthex)
	    _alphabet=(0 1 2 3 4 5 6 7 8 9 A B C D E F G H I J K L M N \
		O P Q R S T U V);;
	bech32)
	    _alphabet=(q p z r y 9 x 8 g f 2 t v d w 0 s 3 j n 5 4 k h c \
		e 6 m u a 7 l);;
	*) return 4
    esac


    # make sure we have even length of hex input string
    local _in_hex_len="${#_in_hex}"
    (( _in_hex_len % 2 == 0 )) || return 92
    _in_hex_len=$((_in_hex_len / 2))

    local _j _i _iset_len _oset_len _result
    local -a _iset _oset
    local _oset_len_lookup=(0 2 4 5 7 8) # index is _iset_len

    _result=
    for (( _j=0; _j < _in_hex_len; _j=_j+5 ))
    do
	# check input set length and set output set length (w/o padding)
	if (( _j + 4 < _in_hex_len ))
	then
	    _iset_len=5
	    _oset_len=8
	else
	    _iset_len=$((_in_hex_len - _j))
	    _oset_len=$((_oset_len_lookup[_iset_len]))
	fi

	# read input set from hex representation
        _iset=(0 0 0 0 0)
	for (( _i=0; _i < _iset_len; _i++ ))
	do
	    _iset[$_i]=$((16#${_in_hex:$(( (_j + _i ) * 2 )):2}))
	done

	# do bit translation
	_oset[0]="$(( _iset[0] >> 3 ))"
	_oset[1]="$(( (_iset[0] & 2#00000111) << 2 | _iset[1] >> 6 ))"
	_oset[2]="$(( (_iset[1] & 2#00111111 ) >> 1 ))"
	_oset[3]="$(( (_iset[1] & 2#00000001) << 4 | _iset[2] >> 4 ))"
	_oset[4]="$(( (_iset[2] & 2#00001111) << 1 | _iset[3] & 1 ))"
	_oset[5]="$(( (_iset[3] & 2#01111111 ) >> 2 ))"
	_oset[6]="$(( (_iset[3] & 2#00000011) << 3 | _iset[4] >> 5 ))"
	_oset[7]="$(( (_iset[4] & 2#00001111) ))"

	# store result and add optional padding
	for (( _i=0; _i < _oset_len; _i++ ))
	do
	    printf -v _result "%s%s" "$_result" "${_alphabet[${_oset[$_i]}]}"
	done
	for (( _i=0; _i < 8 - _oset_len; _i++ ))
	do
	    printf -v _result "%s%s" "$_result" "="
	done
    done

    printf -v "$_ret_vn" "%s" "$_result"
}

# base32_decode <ret-vn> <base32-string> [default|exthex|bech32]
# decodes base32 string to hex representation of binary data
# 3rd argument selects alphabet, see RFC 4648 for details
# https://tools.ietf.org/html/rfc4648
# note: bech32 is untested
# errors: 2, 3, 4, 91 (invalid base32-string length, must be multiple of 8),
#         92 (invalid chars in base32-string)
function bcr_base32_decode_to_hex() {
    (( $# == 2 || $# == 3 )) || {
	bb_ierr "bcr_base32_decode_to_hex called with $#" \
	    "instead of 2 or 3 args"
	return 2
    }

    local -A _rev_alphabet

    local _ret_vn="$1"
    local _base32_str="$2"
    local _mode="${3-default}"

    bb_validate_vn "bb_base32_decode:ret-vn" "$_ret_vn" || return 3

    local _base32_str_len="${#_base32_str}"
    (( _base32_str_len % 8 == 0 )) || return 91

    local -A _rev_alphabet
    case "$_mode" in
	default)
	    [[ "$_base32_str" =~ ^[A-Z2-7=]*$ ]] || return 92
	    _rev_alphabet=( \
		[90]=25 [88]=23 [89]=24 [80]=15 [81]=16 [55]=31 [82]=17 \
		[54]=30 [83]=18 [53]=29 [84]=19 [52]=28 [85]=20 [51]=27 \
		[86]=21 [50]=26 [87]=22 [75]=10 [74]=9 [77]=12 [76]=11 \
		[71]=6 [70]=5 [73]=8 [72]=7 [79]=14 [78]=13 [66]=1 [67]=2 \
		[65]=0 [68]=3 [69]=4 [61]=0)
	    ;;
	exthex)
	    [[ "$_base32_str" =~ ^[0-9A-V=]*$ ]] || return 92
	    _rev_alphabet=( \
		[80]=25 [57]=9 [81]=26 [56]=8 [82]=27 [55]=7 [83]=28 \
		[54]=6 [84]=29 [53]=5 [85]=30 [52]=4 [86]=31 [51]=3 \
		[50]=2 [75]=20 [74]=19 [77]=22 [76]=21 [71]=16 [48]=0 \
		[70]=15 [49]=1 [73]=18 [72]=17 [79]=24 [78]=23 [66]=11 \
		[67]=12 [65]=10 [68]=13 [69]=14 [61]=0)
	    ;;
	bech32)
	    _rev_alphabet=( \
		[99]=24 [97]=29 [109]=27 [108]=31 [101]=25 [57]=5 [100]=13 \
		[56]=7 [55]=30 [103]=8 [54]=26 [102]=9 [53]=20 [104]=23 \
		[52]=21 [107]=22 [51]=17 [106]=18 [50]=10 [116]=11 [117]=28 \
		[114]=3 [115]=16 [48]=15 [112]=1 [113]=0 [110]=19 [118]=12 \
		[119]=14 [122]=2 [121]=4 [120]=6 [61]=0)
	    ;;
	*) return 4
    esac

    local -a _iset _oset
    local _i _j _tmp _oset_len _result

    _result=
    for (( _j=0; _j < _base32_str_len; _j=_j+8 ))
    do
	# check for output set length (input set is always 8 chars)
	if (( _j + 8 >= _base32_str_len ))
	then
	    case "${_base32_str:_j:8}" in
		*======) _oset_len=1;;
		*====) _oset_len=2;;
		*===) _oset_len=3;;
		*=) _oset_len=4;;
		*) _oset_len=5;;
	    esac
	else
	    _oset_len=5
	fi

	# read input from base32 representation
	_iset=(0 0 0 0 0 0 0 0)
	for (( _i=_j; _i < _j + 8 && _i < ${#_base32_str}; _i++ ))
	do
	    printf -v _tmp "%d" "'${_base32_str:_i:1}"
	    _iset[$((_i-_j))]="${_rev_alphabet[$_tmp]}"
	done

	# do bit translation
	_oset[0]="$(( _iset[0] << 3 | _iset[1] >> 2 ))"
	_oset[1]="$(( (_iset[1] & 3) << 6 | _iset[2] << 1 | _iset[3] >> 4 ))"
	_oset[2]="$(( (_iset[3] & 2#00001111) << 4 | _iset[4] >> 1))"
	_oset[3]="$(( (_iset[4] & 1) << 7 | _iset[5] << 2 | _iset[6] >> 3 ))"
	_oset[4]="$(( (_iset[6] & 2#00000111) << 5 | _iset[7] ))"

	# store result without optional padding
	for (( _i=0; _i < _oset_len; _i++ ))
	do
	    printf -v _result '%s%02x' "$_result" "${_oset[$_i]}"
	done
    done

    printf -v "$_ret_vn" "%s" "$_result"
}

#@@ HMAC (RFC 2104)

# shellcheck disable=SC2154
#@ bcr_hmac <ret-vn> <digest> <hex-key> <hex-data>
# where digest is one of: md5|sha(1|224|256|384|512)|ripemd160|whirlpool
# computes hmac defined in https://tools.ietf.org/html/rfc2104
# errors: 2, 3, 4, 91 (illegal char in hex), 92 (hex is of odd length),
#         93 (digest failed)
function bcr_hmac() {
    (( $# == 4 )) || {
	bb_ierr "bcr_mac called with $# instead of 4 args"
	return 2
    }

    local _ret_vn="$1"
    local _digest="$2"
    local bcr_hmac__hex_key="$3"
    local _hex_data="$4"

    bb_validate_vn "bcr_hmac:ret-vn" "$_ret_vn" || return 3
    bb_hex_validate "bcr_hmac:hex-key" "$bcr_hmac__hex_key" "$_hex_data" \
	|| return 91

    (( ${#bcr_hmac__hex_key} % 2 == 0 )) || return 92
    (( ${#_hex_data} % 2 == 0 )) || return 92

    # check digest and set block size
    local _block_size # in number of nibbles (hex string length)
    case "$_digest" in
	md5|sha1|sha224|sha256|ripemd160|whirlpool)
	    _block_size=$((2 * 512 / 8));;
	sha384|sha512)
	    _block_size=$((2 * 1024 / 8));;
	*) return 4
    esac

    # if key is longer than digest block size, apply digest on key and use
    # result as new key
    if [[ ${#bcr_hmac__hex_key} -gt $_block_size ]]; then
	bcr_digest_of_hex bcr_hmac__hex_key "$_digest" \
	    "$bcr_hmac__hex_key" || return 93
    fi

    # make sure _key is padded with zeros up to blocksize
    printf -v bcr_hmac__hex_key "%-${_block_size}s" "$bcr_hmac__hex_key"
    bcr_hmac__hex_key="${bcr_hmac__hex_key// /0}"

    local _ipad _opad
    printf -v _ipad "%${_block_size}s" ""
    _opad="${_ipad//  /5c}"
    _ipad="${_ipad//  /36}"

    # perform the actual hmac operation per FIPS 198-1
    local _keyi bcr_hmac__mac
    bb_hex_xor _keyi "$bcr_hmac__hex_key" "$_ipad"
    bcr_digest_of_hex bcr_hmac__mac "$_digest" "$_keyi$_hex_data" || \
	return 93

    local _keyo
    bb_hex_xor _keyo "$bcr_hmac__hex_key" "$_opad"
    bcr_digest_of_hex bcr_hmac__mac "$_digest" "$_keyo$bcr_hmac__mac" || \
	return 93

    printf -v "$_ret_vn" "%s" "$bcr_hmac__mac"
}

#@@ HOTP (RFC 4226)

#@ bcr_hotp <ret-vn> <hex-key> <counter> <digest> <length>
# calculate HOTP value as specified in RFC 4226
# where counter is a 8 byte integer, digtest is one of sha1, sha256 and sha512
# and length can be 6, 7 or 8 digits
# errors: 2, 3, 4, 91 (HMAC computation failed)
function bcr_hotp() {
    (( $# == 5 )) || {
	bb_ierr "bcr_totp called with $# instead of 5 args"
	return 2
    }

    local _ret_vn="$1"
    local _hex_key="$2"
    local _counter="$3"
    local _digest="${4,,}"
    local _length="$5"

    bb_validate_vn "bcr_hotp:ret-vn" "$_ret_vn" || return 3
    bb_hex_validate "bcr_hotp:hex-key" "$_hex_key" || return 4
    bb_validate_uint "bcr_hotp:counter" "$_counter" || return 4
    bb_validate_regex "bcr_hotp:digest" '^sha(1|256|512)$' "$_digest" || \
	return 4
    bb_validate_regex "bcr_hotp:length" '^[678]$' "$_length" || return 4

    local _hex_counter
    printf -v _hex_counter "%016x" "$_counter"
    (( ${#_hex_counter} < 17 )) || return 4 # counter overflow

    local hotp__hex_hash
    bcr_hmac hotp__hex_hash "$_digest" "$_hex_key" "$_hex_counter" || {
	local _err_code="$?"
	bb_msg err "bcr_hotp: bcr_hmac failed with: $_err_code"
	return 91
    }

    local _sbits
    local _offset
    printf -v _offset "%i" \
	"$(( 16#${hotp__hex_hash:$(( ${#hotp__hex_hash} - 1 )):1} ))"

    local _p _msb
    _p="${hotp__hex_hash:$((_offset * 2)):8}"

    printf -v _msb "%i" "$(( 16#${_p:0:2} ))"
    printf -v _p "%x%s" "$(( _msb & 2#01111111 ))" "${_p:2:6}"

    printf -v _sbits "%s" "$_p"

    local _untruncated
    printf -v _untruncated "%0${_length}i" "$(( 16#$_sbits ))"
    printf -v "$_ret_vn" "%s" "${_untruncated:(-$_length)}"
}

#@@ TOTP (RFC 6238)

#@ totp <ret-vn> <hex-key> <sec-time> <totp_x> <digest> <ret-vn-length>
# calculate TOTP value as specified in RFC 6238
# errors: 2, 3, 4
function bcr_totp() {
    (( $# == 6 )) || {
	bb_ierr "bcr_totp called with $# instead of 6 args"
	return 2
    }

    local _ret_vn="$1"
    local _hex_key="$2"
    local _sec_time="$3"
    local _totp_x="$4"
    local _digest="$5"
    local _length="$6"

    bb_validate_vn "bb_totp%1" "$_ret_vn" || return 3
    bb_hex_validate "bb_totp%2" "$_hex_key" || return 4
    bb_validate_uint "bb_totp%3_4" "$_sec_time" "$_totp_x" || return 4
    bb_validate_regex "bb_totp%5" '^sha(1|256|512)$' "$_digest" || return 4
    bb_validate_regex "bb_totp%6" '^[678]$' "$_length" || return 4

    bcr_hotp "$_ret_vn" "$_hex_key" "$(( _sec_time / _totp_x ))" \
	"$_digest" "$_length"
}
